package eu.ztsh.lfr.core.impl.files;

import eu.ztsh.lfr.config.DataProperties;
import eu.ztsh.lfr.model.Averages;
import eu.ztsh.lfr.model.DataRow;
import eu.ztsh.lfr.model.FileProcessingException;
import eu.ztsh.lfr.model.Temperatures;
import eu.ztsh.lfr.model.events.FileModifiedEvent;
import eu.ztsh.lfr.model.events.FileProcessedEvent;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.context.event.ApplicationReadyEvent;
import org.springframework.context.ApplicationEvent;
import org.springframework.context.ApplicationEventPublisher;
import org.springframework.context.event.EventListener;
import org.springframework.scheduling.annotation.Async;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Service;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Predicate;
import java.util.stream.Stream;

@Service
public class FileLoadingService {

    private static final Logger log = LoggerFactory.getLogger(FileLoadingService.class);
    private final DataProperties dataProperties;
    private final ApplicationEventPublisher eventPublisher;

    private final ConcurrentLinkedQueue<ApplicationEvent> queue = new ConcurrentLinkedQueue<>();
    private final AtomicBoolean working = new AtomicBoolean();

    @Autowired
    public FileLoadingService(DataProperties dataProperties, ApplicationEventPublisher eventPublisher) {
        this.dataProperties = dataProperties;
        this.eventPublisher = eventPublisher;
    }

    @Async
    @EventListener({FileModifiedEvent.class, ApplicationReadyEvent.class})
    public void onFileModified(ApplicationEvent event) {
        queue.add(event);
    }

    @Scheduled(fixedRate=500)
    void runTask() {
        if (working.get()) {
            return;
        }
        ApplicationEvent event;
        log.debug("{} events queued", queue.size());
        do {
            // process only last
            event = queue.poll();
        } while (!queue.isEmpty());
        if (event != null) {
            log.info("Processing event: {}", event.getClass().getSimpleName());
            working.set(true);
            var result = processFile();
            eventPublisher.publishEvent(new FileProcessedEvent(result));
            working.set(false);
        }
    }

    private Averages processFile() {
        var start = System.currentTimeMillis();
        log.debug("Processing file");
        final Temperatures model = new Temperatures();
        try (final Stream<String> lines = Files.lines(Paths.get(dataProperties.fileUrl()))) {
            lines.parallel()
                .filter(Predicate.not(String::isBlank))
                .map(line -> {
                    try {
                        var parts = line.split(";");
                        return new DataRow(parts[0], Integer.parseInt(parts[1].split("-")[0]), Double.parseDouble(parts[2]));
                    } catch (Exception e) {
                        throw new FileProcessingException("Error in line %s".formatted(line), e);
                    }
                })
                .forEach(model::addData);
        } catch (IOException e) {
            throw new FileProcessingException("File error", e);
        }
        var millis = System.currentTimeMillis() - start;
        log.atInfo().setMessage("Elapsed time: {}:{}").addArgument(TimeUnit.MILLISECONDS.toMinutes(millis))
            .addArgument(() -> String.format("%02d", TimeUnit.MILLISECONDS.toSeconds(millis) % 60))
            .log();
        return model.getAverages();
    }

}
