package eu.ztsh.lfr.core.impl.files;

import eu.ztsh.lfr.config.DataProperties;
import eu.ztsh.lfr.model.FileProcessingException;
import eu.ztsh.lfr.model.events.FileProcessedEvent;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Mockito;
import org.springframework.context.ApplicationEvent;
import org.springframework.context.ApplicationEventPublisher;

import java.io.IOException;

import static org.assertj.core.api.Assertions.assertThat;
import static org.assertj.core.api.Assertions.assertThatNoException;
import static org.assertj.core.api.Assertions.assertThatThrownBy;

class FileLoadingServiceTest {

    private final ApplicationEventPublisher eventPublisher = Mockito.mock(ApplicationEventPublisher.class);

    @Test
    void successfulRunTest() {
        var captor = ArgumentCaptor.forClass(FileProcessedEvent.class);
        var service = createFileLoadingService("success.csv");
        service.onFileModified(new TestEvent());
        assertThatNoException().isThrownBy(service::runTask);
        Mockito.verify(eventPublisher, Mockito.times(1))
            .publishEvent(captor.capture());
        var averages = captor.getValue().averages();
        assertThat(averages).hasSize(1);
        var list = averages.get("Warszawa");
        assertThat(list).hasSize(1);
        assertThat(list.getFirst()).satisfies(item -> {
            assertThat(item.getYear()).isEqualTo("2018");
            // 29.2775, but it is BigDecimal with scale 2 and round half up
            assertThat(item.getAverageTemperature().doubleValue()).isEqualTo(29.28);
        });
    }

    @Test
    void notExistingFileTest() {
        var service = createFileLoadingService("not-existing.csv");
        service.onFileModified(new TestEvent());
        assertThatThrownBy(service::runTask)
            .isInstanceOf(FileProcessingException.class)
            .hasRootCauseInstanceOf(IOException.class);
        Mockito.verifyNoInteractions(eventPublisher);
    }

    @Test
    void malformedFileTest() {
        var service = createFileLoadingService("malformed.csv");
        service.onFileModified(new TestEvent());
        assertThatThrownBy(service::runTask)
            .isInstanceOf(FileProcessingException.class)
            .hasMessage("Error in line Warszawa;2018-09-21 11:17:42.31833.27")
            .hasRootCauseInstanceOf(ArrayIndexOutOfBoundsException.class);
        Mockito.verifyNoInteractions(eventPublisher);
    }

    private FileLoadingService createFileLoadingService(String path) {
        return new FileLoadingService(new DataProperties("src/test/resources/" + path), eventPublisher);
    }

    private static class TestEvent extends ApplicationEvent {

        public TestEvent() {
            super("TestEvent");
        }

    }

}
